#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <cfloat>
#include <cmath>

#include "Common.h"
#include "SpecialFunctions.h"

#include <gsl/gsl_sf_gamma.h>

double BesselJ0(double x) {
    return j0(x);
}

double BesselJ1(double x) {
    return j1(x);
}

double BesselJn(int n, double x) {
    return jn(n, x);
}

double SphericalBesselJ0(double x) {
    if(fabs(x) < 1e-4)
        return 1 - pow2(x)/6 + pow4(x)/120 - pow6(x)/5040;
    else
        return sin(x)/x;
}

double SphericalBesselJ1(double x) {
    if(fabs(x) < 1e-4)
        return x/3 - pow3(x)/30 + pow5(x)/840 - pow7(x)/45360;
    else
        return (sin(x) - x*cos(x))/pow2(x);
}

double SphericalBesselJ2(double x) {
    if(fabs(x) < 1e-4)
        return pow2(x)/15 - pow4(x)/210 + pow6(x)/7560;
    else
        return ((3 - pow2(x))*sin(x) - 3*x*cos(x))/pow3(x);
}

double SphericalBesselJ3(double x) {
    if(fabs(x) < 1e-4)
        return pow3(x)/105 - pow5(x)/1890 + pow7(x)/83160;
    else
        return ((15 - 6*pow2(x))*sin(x) - (15*x - pow3(x))*cos(x))/pow4(x);
}

double SphericalBesselJ4(double x) {
    if(fabs(x) < 1e-4)
        return pow4(x)/945 - pow6(x)/20790;
    else
        return ((105 - 45*pow2(x) + pow4(x))*sin(x) - (105*x + 10*pow3(x))*cos(x))/pow5(x);
}

/* Based on the Netlib routine (D)GAMMA by W. J. Cody and L. Stoltz.  Original
 * source and documentation available at
 *   http://netlib.org/specfun/gamma */
double Gamma(double x) {
    /* Constants */
    const double sqrtpi = 0.9189385332046727417803297;
    const double pi = 3.1415926535897932384626434;

    /* Numerator and denominator coefficients for rational minimax approximation over (1,2). */
    const double P[] = { -1.71618513886549492533811e0, 2.47656508055759199108314e1, -3.79804256470945635097577e2, 6.29331155312818442661052e2, 8.66966202790413211295064e2, -3.14512729688483675254357e4, -3.61444134186911729807069e4, 6.64561438202405440627855e4 };
    const double Q[] = { -3.08402300119738975254353e1, 3.15350626979604161529144e2, -1.01515636749021914166146e3, -3.10777167157231109440444e3, 2.25381184209801510330112e4, 4.75584627752788110767815e3, -1.34659959864969306392456e5, -1.15132259675553483497211e5 };

    /* Coefficients for minimax approximation over (12,infty). */
    const double C[] = { -1.910444077728e-3,8.4171387781295e-4, -5.952379913043012e-4, 7.93650793500350248e-4, -2.777777777777681622553e-3, 8.333333333333333331554247e-2, 5.7083835261e-3 };

    /* Machine dependent parameters (reasonable values here) */
    const double xbig = 171.624;
    const double xminin = 2.23e-308;
    const double eps = 2.22e-16;
    const double xinf = 1.79e308;

    double fact, res, y;
    int i;

    fact = 1;
    y = x;

    if(x <= 0) {
        y = -x;
        if(y - (int)y == 0) {
            /* x is a negative integer */
            return xinf;
        }
        else {
            /* Use the reflection formula Gamma(1-x) Gamma(x) = pi/sin(pi x) */
            fact = pi/sin(pi*x);
            y = 1 - x;
        }
    }

    /* y is now positive, and we seek Gamma(y) */

    if(y < eps) {
        /* 0 < y < eps: use limiting formula Gamma(y) -> 1/y */
        if(y >= xminin)
            res = 1/y;
        else
            res = xinf;
    }
    else if(y < 12) {
        int n = ((int)y) - 1;
        double z = y - (n + 1);
        double xnum, xden;

        /* Evaluate minimax approximation for Gamma(1+z) with 0 < z < 1 */
        xnum = 0;
        xden = 1;
        for(i = 0; i < 8; i++) {
            xnum = (xnum + P[i])*z;
            xden = xden*z + Q[i];
        }
        res = 1 + xnum/xden;

        /* Adjust result for y < 1 or y > 2*/
        if(n == -1)
            res /= y;
        else
            for(i = 0; i < n; i++)
                res *= (z + 1 + i);

    }
    else {
        /* y >= 12: use asymptotic expansion */
        if(y > xbig)
            res = xinf;
        else {
            double ysq = y*y;
            double sum = C[6];
            for(i = 0; i < 6; i++)
                sum = sum/ysq + C[i];
            sum = sum/y - y + sqrtpi + (y-0.5)*log(y);
            res = exp(sum);
        }
    }

    if(fact != 1)
        res = fact/res;
    return res;
}

double LowerGamma(double a, double x) {
    if(x < 0.1)
        return pow(x,a)*(1/a - x/(a+1) + 0.5*x*x/(a+2));
    else
        return gsl_sf_gamma(a) - gsl_sf_gamma_inc(a, x);
}

double UpperGamma(double a, double x) {
    return gsl_sf_gamma_inc(a, x);
}
