#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <cassert>
#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <cstring>

#include "array.h"

/* Ideas for refactor:
 * - Treat 1-D, 2-D, and 3_d arrays uniformly, and rely on the user to use them
 *   properly:
 *     add member variable size_type dim[3], with size() == dim[0]*dim[1]*dim[2]
 *     array(n1)          results in dim[0] = n1, dim[1] = dim[2] = 0
 *     array(n1, n2)      results in dim[0] = n1, dim[1] = n2, dim[2] = 0
 *     array(n1, n2, n3)  results in dim[0] = n1, dim[1] = n2, dim[2] = n3
 *     operator()(int i)  returns (*this)[i], i.e. assumes 1-D array
 *     operator()(int i, int j)  returns (*this)[dim[1]*i + j], i.e. assumes 2-D array
 *     operator()(int i, int j, int k)  returns (*this)[dim[2]*(dim[1]*i + j) + k], i.e. assumes 3-D array
 * - Use a template for the type, i.e. array would become array<real>. */


array::array() {
}

array::array(size_type n) : vector<real>(n) {
}

array::array(size_type n, real r) : vector<real>(n, r) {
}

array::array(size_type n, real min, real max, SpacingMode mode) : vector<real>(n) {
    assert(n >= 2);
    if(mode == LinearSpacing) {
        for(size_type i = 0; i < n; i++)
            (*this)[i] = min + i*(max - min)/(n-1);
    }
    else {
        assert(min > 0 && max > 0);
        real logmin = log(min);
        real logmax = log(max);
        for(size_type i = 0; i < n; i++)
            (*this)[i] = exp(logmin + i*(logmax - logmin)/(n-1));
    }

    /* Guarantee min and max elements are exact */
    (*this)[0] = min;
    (*this)[n-1] = max;
}

array::array(size_type n, const real* v) : vector<real>(v, v+n) {
}

array::array(const vector<real>& v) : vector<real>(v) {
}

array& array::operator=(const vector<real>& v) {
    vector<real>::operator=(v);
    return *this;
}

array::~array() {
}

array array::linspace(real min, real max, size_type n) {
    return array(n, min, max, LinearSpacing);
}

array array::logspace(real min, real max, size_type n) {
    return array(n, min, max, LogarithmicSpacing);
}

void array::setdata(size_type n, real r) {
    resize(n, r);
}

void array::setdata(size_type n, const real* v) {
    resize(n);
    for(iterator it = begin(); it != end(); it++)
        *it = *v++;
}

array& array::operator+=(const array& v) {
    size_type n = size();
    assert(v.size() == n);
    for(size_type i = 0; i < n; i++)
        (*this)[i] += v[i];
    return *this;
}

array& array::operator-=(const array& v) {
    size_type n = size();
    assert(v.size() == n);
    for(size_type i = 0; i < n; i++)
        (*this)[i] -= v[i];
    return *this;
}

array& array::operator*=(const array& v) {
    size_type n = size();
    assert(v.size() == n);
    for(size_type i = 0; i < n; i++)
        (*this)[i] *= v[i];
    return *this;
}

array& array::operator/=(const array& v) {
    size_type n = size();
    assert(v.size() == n);
    for(size_type i = 0; i < n; i++)
        (*this)[i] /= v[i];
    return *this;
}

array& array::operator*=(real s) {
    size_type n = size();
    for(size_type i = 0; i < n; i++)
        (*this)[i] *= s;
    return *this;
}

array& array::operator/=(real s) {
    size_type n = size();
    for(size_type i = 0; i < n; i++)
        (*this)[i] /= s;
    return *this;
}

array operator-(const array& v) {
    array::size_type n = v.size();
    array w(n);
    for(array::size_type i = 0; i < n; i++)
        w[i] = -v[i];
    return w;
}

array operator+(const array& u, const array& v) {
    assert(u.size() == v.size());
    array w = u;
    w += v;
    return w;
}

array operator-(const array& u, const array& v) {
    assert(u.size() == v.size());
    array w = u;
    w -= v;
    return w;
}

array operator*(const array& u, const array& v) {
    assert(u.size() == v.size());
    array w = u;
    w *= v;
    return w;
}

array operator/(const array& u, const array& v) {
    assert(u.size() == v.size());
    array w = u;
    w /= v;
    return w;
}

array operator*(const array& u, real s) {
    array w = u;
    w *= s;
    return w;
}

array operator*(real s, const array& u) {
    array w = u;
    w *= s;
    return w;
}

array operator/(const array& u, real s) {
    array w = u;
    w /= s;
    return w;
}

array2::array2() : array() {
    n1 = n2 = 0;
}

array2::array2(size_type n1_, size_type n2_) : array(n1_*n2_) {
    n1 = n1_;
    n2 = n2_;
}

array2::array2(size_type n1_, size_type n2_, real r) : array(n1_*n2_, r) {
    n1 = n1_;
    n2 = n2_;
}

array2::array2(const array2& a) : array(a) {
    n1 = a.n1;
    n2 = a.n2;
}

array2& array2::operator=(const array2& a) {
    array::operator=(a);
    n1 = a.n1;
    n2 = a.n2;
    return *this;
}

void array2::getsize(size_type& nx, size_type& ny) {
    nx = n1;
    ny = n2;
}

void array2::resize(size_type n1_, size_type n2_) {
    n1 = n1_;
    n2 = n2_;
    array::resize(n1*n2);
}

void array2::resize(size_type n1_, size_type n2_, real r) {
    n1 = n1_;
    n2 = n2_;
    array::resize(n1*n2, r);
}

array3::array3() : array() {
    Nx = Ny = Nz = 0;
}

array3::array3(int Nx_, int Ny_, int Nz_) : array(Nx_*Ny_*Nz_) {
    Nx = Nx_;
    Ny = Ny_;
    Nz = Nz_;
}

array3::array3(int Nx_, int Ny_, int Nz_, real r) : array(Nx_*Ny_*Nz_, r) {
    Nx = Nx_;
    Ny = Ny_;
    Nz = Nz_;
}

array3::array3(const array3& a) : array(a) {
    Nx = a.Nx;
    Ny = a.Ny;
    Nz = a.Nz;
}

array3& array3::operator=(const array3& a) {
    array::operator=(a);
    Nx = a.Nx;
    Ny = a.Ny;
    Nz = a.Nz;
    return *this;
}

void array3::resize(int Nx_, int Ny_, int Nz_) {
    Nx = Nx_;
    Ny = Ny_;
    Nz = Nz_;
    array::resize(Nx*Ny*Nz);
}

void array3::resize(int Nx_, int Ny_, int Nz_, real r) {
    Nx = Nx_;
    Ny = Ny_;
    Nz = Nz_;
    array::resize(Nx*Ny*Nz, r);
}
