#ifndef ARRAY_H
#define ARRAY_H

#include <vector>
using std::vector;

#include "Common.h"

enum SpacingMode {
    LinearSpacing,
    LogarithmicSpacing
};

/* Simple 1-dimensional array */
class array : public vector<real> {
public:
    /* Construct an empty array. */
    array();

    /* Construct an array of length n, with arbitrary initial values. */
    explicit array(size_type n);

    /* Construct an array of length n, with all elements set to r. */
    array(size_type n, real r);

    /* Construct an array of length n, with elements evenly spaced between
     * min and max.  (Note that min can be greater than max, if you want the
     * elements to be in descending order.) */
    array(size_type n, real min, real max, SpacingMode mode = LinearSpacing);

    /* Construct an array of length n, with elements initialized from the
     * C-array v.  Note that v must have at least n elements. */
    array(size_type n, const real* v);

    /* Copy constructor. */
    array(const vector<real>& v);

    /* Assignment.  This array is resized, and the elements of v are copied. */
    array& operator=(const vector<real>& v);

    /* Destructor. */
    ~array();

    /* Constuct an array of length n with linearly or logarithmically spaced
     * values between min and max. */
    static array linspace(real min, real max, size_type n = 50);
    static array logspace(real min, real max, size_type n = 50);

    void setdata(size_type n, real r);
    void setdata(size_type n, const real* v);

//    real* data() { return &front(); }
//    const real* data() const { return &front(); }

    const real& operator[](int i) const { return vector<real>::operator[](i); }
    real& operator[](int i) { return vector<real>::operator[](i); }

    /* Automatic cast to C-array.  Note that the returned pointer is only valid
     * as long as this array is not resized. */
    operator real*() { return data(); }
    operator const real*() const { return data(); }

    /* Arithmetic operations. */
    array& operator+=(const array& v);  // element-wise addition
    array& operator-=(const array& v);  // element-wise subtraction
    array& operator*=(const array& v);  // element-wise multiplication
    array& operator/=(const array& v);  // element-wise division
    array& operator*=(real s);          // multiplication by a constant
    array& operator/=(real s);          // division by a constant
};

array operator-(const array& v);    // negation
array operator+(const array& u, const array& v);
array operator-(const array& u, const array& v);
array operator*(const array& u, const array& v);
array operator/(const array& u, const array& v);
array operator*(const array& u, real s);
array operator*(real s, const array& u);
array operator/(const array& u, real s);


/* Simple 2-dimensional array */
class array2 : public array {
public:
    array2();
    array2(size_type n1, size_type n2);
    array2(size_type n1, size_type n2, real r);
    array2(const array2& a);
    array2& operator=(const array2& a);

    void getsize(size_type& n1, size_type& n2);
    void resize(size_type n1, size_type n2);
    void resize(size_type n1, size_type n2, real r);

    real operator()(int i, int j) const { return (*this)[i*n2 + j]; }
    real& operator()(int i, int j) { return (*this)[i*n2 + j]; }

protected:
    size_type n1, n2;
};


/* Simple 3-dimensional array */
class array3 : public array {
public:
    array3();
    array3(int Nx, int Ny, int Nz);
    array3(int Nx, int Ny, int Nz, real r);
    array3(const array3& a);
    array3& operator=(const array3& a);

    void resize(int Nx, int Ny, int Nz);
    void resize(int Nx, int Ny, int Nz, real r);

    real operator()(int i, int j, int k) const { return (*this)[i*Ny*Nz + j*Nz + k]; }
    real& operator()(int i, int j, int k) { return (*this)[i*Ny*Nz + j*Nz + k]; }

protected:
    int Nx, Ny, Nz;
};

#endif // ARRAY_H
